{
    "namespace": "suzieq",
    "name": "routes",
    "type": "record",
    "fields": [
        {
	    "name": "sqvers",
	    "type": "string",
	    "partition": 0,
	    "default": "2.0"
	},
        {
            "name": "weights",
            "type": {
                "type": "array",
                "items": {
                    "name": "weight",
                    "type": "long"
                }
            }
        },
        {
            "name": "prefix",
            "type": "string",
            "key": 3,
            "display": 3
        },
        {
            "name": "prefixlen",
            "type": "long",
	    "depends": "prefix"
        },
        {
            "name": "numNexthops",
            "type": "long",
	    "depends": "oifs nexthopIps"
        },
        {
            "name": "source",
            "type": "string",
            "display": 7
        },
        {
            "name": "action",
            "type": "string",
            "display": 10
        },
        {
            "name": "routeTag",
            "type": "string"
        },
        {
            "name": "nexthopIps",
            "type": {
                "type": "array",
                "items": {
                    "name": "nexthop",
                    "type": "string"
                }
            },
            "display": 4
        },
        {
            "name": "asPathList",
            "type": {
                "type": "array",
                "items": {
                    "name": "asPath",
                    "type": "string"
                }
            }
        },
        {
            "name": "oifs",
            "type": {
                "type": "array",
                "items": {
                    "name": "oif",
                    "type": "string"
                }
            },
            "display": 5
        },
        {
            "name": "vrf",
            "type": "string",
            "key": 2,
            "display": 2
        },
        {
            "name": "validState",
            "type": "string"
        },
        {
            "name": "ipvers",
            "type": "long",
            "display": 9
        },
        {
            "name": "protocol",
            "type": "string",
            "display": 6
        },
        {
            "name": "hardwareProgrammed",
            "type": "string"
        },
        {
            "name": "metric",
            "type": "long"
        },
	{
            "name": "statusChangeTimestamp",
            "type": "timestamp",
	    "description": "The last time this route changed"
	},
        {
            "name": "preference",
            "type": "long",
            "display": 8
        },
        {
            "name": "hostname",
            "type": "string",
            "key": 1,
            "display": 1,
	    "partition": 2
        },
        {
            "name": "namespace",
            "type": "string",
            "key": 0,
            "display": 0,
	    "partition": 1
        },
        {
            "name": "timestamp",
            "type": "timestamp"
        },
        {
            "name": "active",
            "type": "boolean"
        }
    ]
}
