{
    "namespace": "suzieq",
    "name": "sqPoller",
    "type": "counter",
    "fields": [
        {
	    "name": "sqvers",
	    "type": "string",
	    "partition": 0,
	    "default": "2.0"
	},
        {
            "name": "gatherTime",
            "type": {
                "type": "array",
                "items": {
                    "type": "float",
                    "name": "time"
                }
            },
            "display": 4	    
        },
        {
            "name": "totalTime",
            "type": {
                "type": "array",
                "items": {
                    "type": "float",
                    "name": "time"
                }
	    },	
            "display": 5	    
        },
        {
            "name": "svcQsize",
            "type": {
                "type": "array",
                "items": {
                    "type": "float",
                    "name": "svcQsize"
                }
	    },
            "display": 6
        },
        {
            "name": "wrQsize",
            "type": {
                "type": "array",
                "items": {
                    "type": "float",
                    "name": "wrQsize"
                }
	    },
            "display": 7
        },
        {
            "name": "nodeQsize",
            "type": {
                "type": "array",
                "items": {
                    "type": "float",
                    "name": "nodeQsize"
                }
	    },
            "display": 8
        },
        {
            "name": "rxBytes",
            "type": {
                "type": "array",
                "items": {
                    "type": "float",
                    "name": "rxBytes"
                }
	    },
            "display": 9
        },
        {
            "name": "version",
            "type": "string"
        },
        {
            "name": "pollExcdPeriodCount",
            "type": "long",
	    "display": 10
        },
        {
            "name": "status",
            "type": "long",
            "display": 3
        },
        {
            "name": "nodesPolledCnt",
            "type": "long"
        },
        {
            "name": "nodesFailedCnt",
            "type": "long"
        },
        {
            "name": "service",
            "type": "string",
            "key": 2,
            "display": 2,
	    "partition": 2
        },
        {
            "name": "hostname",
            "type": "string",
            "key": 1,
            "display": 1,
	    "partition": 2
        },
        {
            "name": "namespace",
            "type": "string",
            "key": 0,
	    "display": 0,         
	    "partition": 1
        },
        {
            "name": "timestamp",
            "type": "timestamp"
        },
        {
            "name": "active",
            "type": "boolean"
        }
    ],
    "recordType": "counters"
}