from os.path import dirname, basename, isfile, join
import glob

name = "pages"

modules = glob.glob(join(dirname(__file__), "*.py"))
__all__ = [basename(f)[:-3] for f in modules if isfile(f)
           and not f.endswith('__init__.py')]

sq_gui_pages_all = __all__
