{
    "namespace": "suzieq",
    "name": "ifCounters",
    "type": "record",
    "fields": [
        {
	    "name": "sqvers",
	    "type": "string",
	    "partition": 0,
	    "default": "1.0"
	},
        {
            "name": "txDrops",
            "type": "long"
        },
        {
            "name": "txCollisions",
            "type": "long"
        },
        {
            "name": "ifname",
            "type": "string",
            "key": 2,
            "display": 2
        },
        {
            "name": "txErrs",
            "type": "long"
        },
        {
            "name": "txPackets",
            "type": "long"
        },
        {
            "name": "rxErrs",
            "type": "long"
        },
        {
            "name": "rxFifo",
            "type": "long"
        },
        {
            "name": "txCarrier",
            "type": "long"
        },
        {
            "name": "rxMcst",
            "type": "long"
        },
        {
            "name": "txBytes",
            "type": "long"
        },
        {
            "name": "txFifo",
            "type": "long"
        },
        {
            "name": "rxBytes",
            "type": "long"
        },
        {
            "name": "rxDrops",
            "type": "long"
        },
        {
            "name": "rxPackets",
            "type": "long"
        },
        {
            "name": "txCompressed",
            "type": "long"
        },
        {
            "name": "rxCompressed",
            "type": "long"
        },
        {
            "name": "rxFrame",
            "type": "long"
        },
        {
            "name": "hostname",
            "type": "string",
            "key": 1,
            "display": 1,
	    "partition": 2
        },
        {
            "name": "namespace",
            "type": "string",
            "key": 0,
            "display": 0,
	    "partition": 1
        },
        {
            "name": "timestamp",
            "type": "timestamp"
        },
        {
            "name": "active",
            "type": "boolean"
        }
    ],
    "recordType": "counters"
}