{
    "namespace": "suzieq",
    "name": "inventory",
    "type": "record",
    "fields": [
        {
            "name": "sqvers",
            "type": "string",
            "partition": 0,
            "default": "1.0"
        },
        {
            "name": "name",
            "type": "string",
            "key": 2,
            "display": 2
        },
        {
            "name": "model",
            "type": "string",
            "key": 3,
            "display": 3
        },
        {
            "name": "status",
            "type": "string",
            "display": 4
        },
        {
            "name": "serial",
            "type": "string",
            "key": 4
        },
        {
            "name": "partNum",
            "type": "string",
            "key": 5
        },
        {
            "name": "vendor",
            "type": "string",
            "display": 5
        },
        {
            "name": "type",
            "type": "string",
            "display": 6
        },
        {
            "name": "version",
            "type": "string",
            "display": 7
        },
        {
            "name": "hostname",
            "type": "string",
            "key": 1,
            "display": 1,
            "partition": 2
        },
        {
            "name": "namespace",
            "type": "string",
            "key": 0,
            "display": 0,
            "partition": 1
        },
        {
            "name": "timestamp",
            "type": "timestamp"
        },
        {
            "name": "active",
            "type": "boolean"
        }
    ]
}