{
    "namespace": "suzieq",
    "name": "ospf",
    "type": "derivedRecord",
    "depends": [
           "ospfNbr",
	   "ospfIf"
    ],
    "fields": [
        {
	    "name": "sqvers",
	    "type": "string",
	    "partition": 0,
	    "default": "1.0"
	},
        {
            "name": "lastDownTime",
            "type": "timestamp"
        },
        {
            "name": "area",
            "type": "string",
            "display": 5
        },
        {
            "name": "nbrPrio",
            "type": "long"
        },
        {
            "name": "lastUpTime",
            "type": "timestamp"
        },
        {
            "name": "instance",
            "type": "string"
        },
        {
            "name": "reason",
            "type": "string"
        },
        {
            "name": "areaStub",
            "type": "boolean"
        },
        {
            "name": "adjState",
            "type": "string",
            "display": 8
        },
        {
            "name": "peerIP",
            "type": "string",
            "display": 9
        },
        {
            "name": "bfdStatus",
            "type": "string"
        },
        {
            "name": "numChanges",
            "type": "long",
            "display": 10
        },
        {
            "name": "lastChangeTime",
            "type": "timestamp",
            "display": 11
        },
        {
            "name": "lsaRetxCnt",
            "type": "long"
        },
        {
            "name": "ifname",
            "type": "string",
            "key": 3,
            "display": 3
        },
        {
            "name": "peerHostname",
            "type": "string",
            "display": 4,
	    "depends": "hostname vrf peerIP ipAddress"
        },
        {
            "name": "vrf",
            "type": "string",
            "key": 2,
            "display": 2
        },
        {
            "name": "peerRouterId",
            "type": "string"
        },
        {
            "name": "cost",
            "type": "long"
        },
        {
            "name": "retxTime",
            "type": "long"
        },
        {
            "name": "deadTime",
            "type": "long"
        },
        {
            "name": "helloTime",
            "type": "long"
        },
        {
            "name": "networkType",
            "type": "string"
        },
        {
            "name": "nbrCount",
            "type": "long",
            "display": 7
        },
        {
            "name": "maskLen",
            "type": "long"
        },
        {
            "name": "isUnnumbered",
            "type": "boolean"
        },
        {
            "name": "ifState",
            "type": "string",
            "display": 6
        },
        {
            "name": "ipAddress",
            "type": "string"
        },
        {
            "name": "routerId",
            "type": "string"
        },
        {
            "name": "hostname",
            "type": "string",
            "key": 1,
            "display": 1,
	    "partition": 2
        },
        {
            "name": "namespace",
            "type": "string",
            "key": 0,
            "display": 0,
	    "partition": 1
        },
        {
            "name": "timestamp",
            "type": "timestamp"
        },
        {
            "name": "active",
            "type": "boolean"
        }
    ]
}