{
    "namespace": "suzieq",
    "name": "lldp",
    "type": "record",
    "fields": [
        {
	    "name": "sqvers",
	    "type": "string",
	    "partition": 0,
	    "default": "1.0"
	},
        {
            "name": "description",
            "type": "string",
            "display": 5
        },
        {
            "name": "peerIfname",
            "type": "string",
            "display": 4
        },
        {
            "name": "peerIfindex",
            "type": "long"
        },
        {
            "name": "peerMacaddr",
            "type": "string"
        },
        {
            "name": "mgmtIP",
            "type": "string",
            "display": 6
        },
        {
            "name": "peerHostname",
            "type": "string",
            "display": 3
        },
        {
            "name": "subtype",
            "type": "string"
        },
        {
            "name": "ifname",
            "type": "string",
            "key": 2,
            "display": 2
        },
        {
            "name": "hostname",
            "type": "string",
            "key": 1,
            "display": 1,
	    "partition": 2
        },
        {
            "name": "namespace",
            "type": "string",
            "key": 0,
            "display": 0,
	    "partition": 1
        },
        {
            "name": "timestamp",
            "type": "timestamp"
        },
        {
            "name": "active",
            "type": "boolean"
        }
    ]
}