{
    "namespace": "suzieq",
    "name": "topology",
    "type": "derivedRecord",
    "depends": [
        "address",
        "arpnd",
        "bgp",
        "device",
        "evpnVni",
        "interfaces",
        "lldp", 
        "macs",
        "ospf",
        "routes"
    ],
    "fields": [
        {
	    "name": "sqvers",
	    "type": "string",
	    "partition": 0,
	    "default": "1.0"
	},

        {
            "name": "namespace",
            "type": "string",
            "display": 1,
            "description": "namespace within which path analysis is run",
            "partition": 1
        },
        {
            "name": "hostname",
            "type": "string",
            "display": 2,
            "description": "name of the host being queried",
            "partition": 2
        },
        {
            "name": "peerHostname",
            "type": "string",
            "display": 4,
            "description": "name of the other side of the connection"
        },
        {
            "name": "ifname",
            "type": "string",
            "display": 3,
            "description": "interface that the host is connected on, discovered through lldp"

        },
        {
            "name": "peerIP",
            "type": "string",
            "display": 5,
            "description": "IP Address of peer, discovered through ARP"

        },
        {
            "name": "vrf",
            "type": "string",
            "display": 6,
            "description": "VRF that the connection is in"
        },
        {
            "name": "arpnd",
            "type": "bool",
            "display": 7,
            "description": "is this connection discovered via ARP"
        },
        {
            "name": "arpndBidir",
            "type": "bool",
            "display": 8,
            "description": "is the ARP entry bidirecttonal"
        },
        {
            "name": "bgp",
            "type": "bool",
            "display": 9,
            "description": "is this connection discovered via BGP"
        },
        {
            "name": "lldp",
            "type": "bool",
            "display": 10,
            "description": "is this connection discovered via LLDP"
        },
        {
            "name": "ospf",
            "type": "bool",
            "display": 11,
            "description": "is this connection discovered via OSPF"
        },
        {
            "name": "polled",
            "type": "bool",
            "display": 12,
            "description": "is this peerHostname polled by Suzieq"
        }
    ]
}